#!/usr/bin/perl -w
#
# diffFinder3.pl
# --------------
#
# Copyright (C) 2001 Washington University School of Medicine
# and Howard Hughes Medical Institute
# All rights reserved
#
# Author: Christian M. Zmasek 
#         zmasek@genetics.wustl.edu
#         http://www.genetics.wustl.edu/eddy/people/zmasek/
#
# Created: 08/17/01
#
# Last modified 08/28/01
#
# Usage: "diffFinder3.pl <infile = Xrio.pl-output> <outfile>
#        <max ratio dists non-ortho/ortho (<1.0)>
#        <X = min bootstraps for being orthologous>
#        <Y = min bootstraps being checked for ratio>
#        <Z = max bootstraps for not being orthologous> [species to ignore]
#
# Counts and stores the RIO analyzed query/domains for which:
# 1. ALL sequences with orthology bootstrap values higher or equal to Y have
#    larger pairwise distances (between the query and the individual hit)
#    than at least one non-ortholog (<=Z).
# 2. There is a least one ortholog with orthology bootstrap values higher
#    or equal to X.
#
# Important: The result of this is meaningful ONLY if the thresholds
# for output of the RIO analysis are set to zero (L=0 R=0) and if
# sorted according to orthology values as highest priority.
#
# Format for infile: 
# ... 
# 
# # ############################################################################
# # Annotation: 2L52.1 CE20433    () TR:Q9XWB3 protein_id:CAA21776.1
# # HMM       : zf-C2H2
# # score     : 17.4
# # E-value   : 0.016
# # No PWD file for this family.
# # ############################################################################
# 
# 
# # ############################################################################
# # Annotation: 4R79.1 CE19649   Zinc-binding metalloprotease domain (HINXTON) TR:Q9U3S9 protein_id:CAB63429.1
# # HMM       : Astacin
# # score     : 296.3
# # E-value   : 1.8e-86
# RIO - Resampled Inference of Orthologs
# Version: 3.020
# ------------------------------------------------------------------------------
# Pfam alignment file                     : /nfs/w8/scratch/zmasek/pfam6.2_tol1-4/ALNs/Astacin.aln
# Bootstrapped pairwise distances file    : /nfs/w8/scratch/zmasek/pfam6.2_tol1-4/PWDs/Astacin.pwd
# Not bootstrapped pairwise distances file: /nfs/w8/scratch/zmasek/pfam6.2_tol1-4/NBDs/Astacin.nbd
# Bootstrap positions file                : /nfs/w8/scratch/zmasek/pfam6.2_tol1-4/BSPs/Astacin.bsp
# HMM                                     : /nfs/w8/scratch/zmasek/pfam6.2_tol1-4/calibratedHMMs/Astacin.hmm
# Query file                              : /tmp/Xriopl9949642330/QUERY
# ==============================================================================
# 
# Query         : CE19649_CAEEL/2-178
# 
# Number (in %) of observed orthologies (o) and super orthologies (s) to query
# in bootstrapped trees, evolutionary distance to query (as average number
# of aa replacements per residue):
# 
# Sequence                Description                                                                o[%] s[%]  distance
# --------                -----------                                                                ---- ----  --------
# O42326_CYPCA/85-272     NEPHROSIN PRECURSOR.                                                         12    0  0.856940
# MEPB_RAT/70-258         MEPRIN A BETA-SUBUNIT PRECURSOR (EC 3.4.24.18) (ENDOPEPTIDASE-2).             7    0  0.904810
# MEPB_HUMAN/69-257       MEPRIN A BETA-SUBUNIT PRECURSOR (EC 3.4.24.18) (ENDOPEPTIDASE-2) (N-BENZO     7    0  0.912330
#                         YL-L-TYROSYL-P-AMINO-BENZOIC ACID HYDROLASE BETA SUBUNIT) (PABAPEPTIDE HY
#                         DROLASE) (PPH BETA).
# MEPB_MOUSE/70-258       MEPRIN A BETA-SUBUNIT PRECURSOR (EC 3.4.24.18) (ENDOPEPTIDASE-2).             7    0  0.933180   
#  ... 
# Q22398_CAEEL/52-200     T11F9.6 PROTEIN.                                                              0    0  1.655080
# 
# 
# Mean bootstrap value of consensus tree (sd): 56% (+/-38%)
# 
# 
# Distance values (based on NJ tree of original alignment)
# --------------------------------------------------------
# Given the thresholds for distance calculations:
# No sequence is considered orthologous to query.
# 
# distance of query to root                     = 0.53786
# mean of distances (for all sequences) to root = 0.678052
# sd of distances (for all sequences) to root   = 0.143856
# n (sum of sequences in alignment plus query)  = 74
# 
# 
# 
# ==============================================================================
# RIO options
# -----------
# Mode                                                      : precalc....
# Bootstraps                                                : 100
# Species tree                                              : /nfs/wol2/...
# Saved annotated NJ tree (based on original alignment)     : /nfs/wol2/p...
# Threshold for output - orthologies               (L=)     : 0
# Threshold for output - super orthologies         (R=)     : 0
# Threshold for distance calc. - orthologies       (U=)     : 80
# Threshold for distance calc. - super orthologies (V=)     : 0
# When to generate warnings:
# More than one ortholog:  diff. in standard deviations (X=): 2
# No  orthologs         :  diff. in standard deviations (Y=): 2
# One ortholog          :  factor                       (Z=): 2
# Sort priority: orthologies > distance to query > super orthologies
# 
# Options for the calculation of the phylgenetic trees
# ----------------------------------------------------
# Model for pairwise distance calculations           : BLOSUM 62 (Henikoff-Henikoff 92)
# Columns in alignment used for tree calc            : 196
# Columns in original alignment                      : 323
# Sequences RIO analyzed                             : 73
# Sequences in alignment used for trees (incl. query): 74
# Program to calculate pairwise distances            : TREE-PUZZLE
# Program to calculate tree                          : PHYLIP NEIGHBOR (NJ)
# Prgrm to calc. branch lengths for distance values  : PHYLIP NEIGHBOR (NJ)
# Seed for random number generator                   : 41
# Options for makeTree                               : XRS41FL
# 
# Time and date
# -------------
# Time requirement to prepare query    : 0 s
# Time requirement to prepare HMM      : 0 s
# Time requirement to align            : 0 s
# Time requirement to get names f. pwd : 0 s
# Time requirement move query to last  : 0 s
# Time requirement rem seqs not in pwd : 0 s
# Time requirement bootstrap           : 2 s
# Time requirement dqo puzzle          : 47 s
# Time requirement add dists           : 1 s
# Time requirement for tree calculation: 5 s
# Time requirement for SDI and RIO     : 8 s
# Total time requirement               : 72 s
# Date started                         : Thu Jul 12 13:57:13 CDT 2001
# Date finished                        : Thu Jul 12 13:58:25 CDT 2001
# 
# Command line
# ------------
# /nfs/wol2/people/zmasek/perl/rio.pl 1 A=Asta......
# # Successful calculations                  : 1
# # No calculation due to absence of PWD file: 1
# # Calculation already performed            : 0
# # ############################################################################
# ....                                                                                                                                     


use strict;

my $VERSION                  = 3.000;

my $infile                   = "";       
my $outfile                  = "";
my $outfile_riooutput        = "";
my $return_line              = "";
my $flag                     = 0;
my $max_ratio                = 1.0;
my $min_bs_for_ortholog      = -1;
my $min_bs_to_check          = -1;
my $max_bs_for_not_orthologs = -1;
my @species_to_ignore        = ();
my $annotation               = "";
my $HMM                      = "";
my $total_domains            = 0;
my $counter                  = 0;
my @o_bootstraps_array       = ();
my @dists_array              = ();
my @found_domains            = ();
my @riooutput                = ();
my $bootstraps               = 0;
my $dist                     = 0;
my $n                        = 0;
my $i                        = 0;
my $j                        = 0;
my $a;
my $b;
my $c;
my $d;



if ( @ARGV < 6 ) {
    &errorInCommandLine();
    exit ( -1 ); 
}

$infile                   = $ARGV[ 0 ];
$outfile                  = $ARGV[ 1 ];
$max_ratio                = $ARGV[ 2 ];
$min_bs_for_ortholog      = $ARGV[ 3 ];
$min_bs_to_check          = $ARGV[ 4 ];
$max_bs_for_not_orthologs = $ARGV[ 5 ];

if ( $max_ratio <= 0.0 
|| $min_bs_for_ortholog < 1
|| $min_bs_to_check < 1
|| $max_bs_for_not_orthologs < 0 
|| ( $min_bs_for_ortholog <= $max_bs_for_not_orthologs ) 
|| ( $min_bs_for_ortholog < $min_bs_to_check )
|| ( $max_bs_for_not_orthologs >= $min_bs_to_check )
|| $max_ratio >= 1.0 
|| $min_bs_for_ortholog > 100
|| $min_bs_to_check > 100
|| $max_bs_for_not_orthologs > 100 ) {
    &errorInCommandLine();
    exit ( -1 );
}


if ( @ARGV >= 7 ) {
    for ( my $x = 6; $x < @ARGV; ++$x ) { 
        $species_to_ignore[ $x - 6 ] = $ARGV[ $x ];
    }
}

$outfile_riooutput = $outfile.".rio.out";

if ( -e $outfile ) {
    die "\n$0: <<$outfile>> already exists.\n";
}
if ( -e $outfile_riooutput ) {
    die "\n$0: <<$outfile_riooutput>> already exists.\n";
}
unless ( ( -s $infile ) && ( -f $infile ) && ( -T $infile ) ) {
    die "\n$0: <<$infile>> does not exist, is empty, or is not a plain textfile.\n";
}

open( OUT_RIOOUT, ">$outfile_riooutput" ) || die "\n$0: Cannot create file \"$outfile_riooutput\": $!\n";

open( IN, "$infile" ) || die "\n$0: Cannot open file <<$infile>>: $!\n";

while ( $return_line = <IN> ) {

    if ( $return_line =~ /#\s+Annotation\s*:\s*(.+)/ ) {
        $annotation = $1;
        @riooutput  = ();
        $n          = 0;
        $riooutput[ $n++ ] = $return_line;
    }
    elsif ( $return_line =~ /#\s+HMM\s*:\s*(.+)/ ) {
        $HMM = $1;
        $HMM =~ /([^\/]+)$/;
        $HMM = $1;
        $riooutput[ $n++ ] = $return_line;
    }
    elsif ( $return_line =~ /^--------\s+/ ) {
        $flag               = 1;
        $i                  = 0; 
        @o_bootstraps_array = ();
        @dists_array        = ();
        $total_domains++;
    }
    elsif ( $return_line =~ /^Mean\s+bootstrap\s+/i ) {
        if ( $annotation eq "" || $HMM eq "" ) {
            print "HMM       : $HMM\n";
            print "Annotation: $annotation\n";
            die "\n\n$0: Error.\n\n";
        } 
        &analyze(); 
        $n = 2;
        
    }
    elsif ( $return_line =~ /^RIO\s+options\s+/i ) {
        $flag = 0;
        $annotation = "";
        $HMM        = "";
    }
    elsif ( $flag == 1 && $return_line =~ /^\S+.+\s+(\d+)\s+\d+\s+(\S+)\s*$/ ) {
        $bootstraps = $1;
        $dist       = $2;
        $riooutput[ $n++ ] = $return_line;
        if ( $bootstraps > 100 || $bootstraps < 0 ) {
            print "O bootstrap value: $bootstraps\n";
            die "\n\n$0: Error: Boostrap value out of range.\n\n";
        }
        if ( $dist > 10 || $dist < 0 ) {
            print "Distance value: $dist\n";
            die "\n\n$0: Error: Distance value out of range.\n\n";
        }

        if ( scalar( @species_to_ignore ) > 0 ) {
            my $flag = 0;
            for ( my $x = 0; $x < scalar( @species_to_ignore ); ++$x ) {
                my $s = $species_to_ignore[ $x ]; 
                if ( $return_line =~ /^\S+_$s\/\S+\s+/ ) {
                    $flag = 1;
                    last;
                }
            }
            if ( $flag == 0 ) {
                &store;
            }
        }
        else {
            &store;
        }
    }
}

close( IN );

close( OUT_RIOOUT );

open( OUT, ">$outfile" ) || die "\n$0: Cannot create file \"$outfile\": $!\n";

print OUT "diffFinder2.pl version: $VERSION\n";
print OUT "--------------\n\n";
print OUT "Counts and stores the RIO analyzed query/domains for which:\n";
print OUT "1. ALL sequences with orthology bootstrap values higher or equal to Y have\n";
print OUT "   larger pairwise distances (between the query and the individual hit)\n";
print OUT "   than at least one non-ortholog (<=Z).\n"; 
print OUT "2. There is a least one ortholog with orthology bootstrap values higher\n";
print OUT "   or equal to X.\n\n";
print OUT "Input file     : $infile\n";
print OUT "Output file    : $outfile\n";
print OUT "RIO output file: $outfile_riooutput\n";
print OUT "Date           : ".`date`."\n\n";
print OUT "Max ratio dists non-ortho/ortho (<1.0)       = $max_ratio\n";
print OUT "Min bootstraps for being orthologous (X)     = $min_bs_for_ortholog\n";
print OUT "Min bootstraps being checked for ratio (Y)   = $min_bs_to_check\n";
print OUT "Max bootstraps for not being orthologous (Z) = $max_bs_for_not_orthologs\n\n";
print OUT "Species to ignore:   @species_to_ignore\n\n";

print OUT "Total of domains analyzed          : $total_domains\n";
print OUT "Total of domains meeting conditions: $counter\n\n";


print OUT "       Annotation                                                                                      HMM                  min ratio # of orthologs\n";

print OUT "       ----------------------------------------------------------------------------------------------- -------------------- --------- ---- \n\n";

format OUT =
@<<<<< @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< @<<<<<<<<<<<<<<<<<< @#.#####   @<<<
$j,    $a,                                                                                             $b,                 $c,        $d
.

for ( $j = 0; $j < $counter; ++$j ) {
    ( $a, $b, $c, $d ) = split( /@@/ , $found_domains[ $j ] );
    write ( OUT );
}
close( OUT );

print( "\nDone.\n\n" );

exit( 0 );



sub store {
    $o_bootstraps_array[ $i ] = $bootstraps;
    $dists_array[ $i ]        = $dist;
    $i++;
}


sub analyze {

    my $x = 0;
    my $y = 0;
    my $bs_diff = 0;
    my $ratio = 0.0;
    my $smallest_ratio = 1.0;
    my $missed_orthologs_per_domain_counter = 0;
    my $saw_non_ortho_with_smaller_ratio = 0;
    my $saw_inbetween_with_smaller_ratio = 0;
    my $saw_big_enough_diff = 0;

    X: for ( $x = 0; $x < $i; ++$x ) {
        if ( $o_bootstraps_array[ $x ] < $min_bs_to_check ) {
            last X;
        }
        if ( $dists_array[ $x ] <= 0.0 ) {
            next X;
        }
        
        $saw_non_ortho_with_smaller_ratio = 0;
        $saw_inbetween_with_smaller_ratio = 0;
        Y: for ( $y = $x + 1; $y < $i; ++$y ) {
            if ( $o_bootstraps_array[ $y ] > $max_bs_for_not_orthologs ) {
                next Y;
            }
            if ( $dists_array[ $y ] <= 0.0 ) {
                next Y;
            }
            $bs_diff = $o_bootstraps_array[ $x ] - $o_bootstraps_array[ $y ];
            if ( $bs_diff < 0 ) {
                print "Diff in bootstrap values: $bs_diff\n";
                die "\n\n$0: Error: Orthologs appear not to be sorted according to orthology.\n\n";
            } 
            $ratio = $dists_array[ $y ] / $dists_array[ $x ];
            
           
            if ( $ratio <= $max_ratio ) {
                if ( $o_bootstraps_array[ $x ] >= $min_bs_for_ortholog ) {
                    $saw_non_ortho_with_smaller_ratio = 1;
                    if ( $ratio < $smallest_ratio ) {
                        $smallest_ratio = $ratio;
                    }
                }
                else {
                    $saw_inbetween_with_smaller_ratio = 1;
                }
            } 
        } # y loop
        
        if ( $saw_non_ortho_with_smaller_ratio == 1 ) {
            $missed_orthologs_per_domain_counter++;
        }
        elsif ( $saw_inbetween_with_smaller_ratio != 1 ) { 
            return;
        }
       
    } # x loop

    if ( $missed_orthologs_per_domain_counter > 0 ) {
        $found_domains[ $counter ] = $annotation."@@".$HMM."@@".$smallest_ratio."@@".$missed_orthologs_per_domain_counter;
        $counter++;
        print "$counter\n";
        for ( $x = 0; $x < $n; $x++ ) {
            print OUT_RIOOUT $riooutput[ $x ];
        }
        print OUT_RIOOUT "\n\n";
    }
    return;
}



sub errorInCommandLine {
    print "\n";
    print " diffFinder3.pl version: $VERSION\n\n";
    print " Usage: \"diffFinder3.pl <infile = Xrio.pl-output> <outfile>\n";
    print "        <max ratio dists non-ortho/ortho (<1.0)>\n";
    print "        <X = min bootstraps for being orthologous>\n";
    print "        <Y = min bootstraps being checked for ratio>\n";
    print "        <Z = max bootstraps for not being orthologous> [species to ignore]\"\n";
    print " Counts and stores the RIO analyzed query/domains for which:\n";
    print " 1. ALL sequences with orthology bootstrap values higher or equal to Y have\n";
    print "    larger pairwise distances (between the query and the individual hit)\n";
    print "    than at least one non-ortholog (<=Z).\n"; 
    print " 2. There is a least one ortholog with orthology bootstrap values higher\n";
    print "    or equal to X.\n\n";
    print " Important: The result of this is meaningful ONLY if the thresholds\n"; 
    print " for output of the RIO analysis are set to zero (L=0 R=0) and if\n";
    print " sorted according to orthology values as highest priority.\n";
    print "\n";
    exit( -1 );
}


